package com.thaiopensource.pullax;

public class XmlFeature {
  protected XmlFeature() { }

  static private final class BuiltinFeature extends XmlFeature {
    private final String name;

    private BuiltinFeature(String name) {
      this.name = name;
    }
    
    public String toString() {
      return name;
    }
  }

  /**
   * The feature that controls handling of empty elements.
   *
   * @see XmlItem#START_ELEMENT
   * @see XmlItem#END_ELEMENT
   * @see XmlItem#EMPTY_ELEMENT
   */
  public static final XmlFeature SHOW_EMPTY_ELEMENT
    = new BuiltinFeature("SHOW_EMPTY_ELEMENT");

  /**
   * The feature that controls handling of CDATA sections.
   *
   * @see XmlItem#CDATA_SECTION
   */
  public static final XmlFeature SHOW_CDATA_SECTION
    = new BuiltinFeature("SHOW_CDATA_SECTION");

  /**
   * The feature that controls handling of processing instructions.
   *
   * @see XmlItem#PROCESSING_INSTRUCTION
   */
  public static final XmlFeature SHOW_PROCESSING_INSTRUCTION
    = new BuiltinFeature("SHOW_PROCESSING_INSTRUCTION");

  /**
   * The feature that controls handling of comments.
   *
   * @see XmlItem#COMMENT
   */
  public static final XmlFeature SHOW_COMMENT
    = new BuiltinFeature("SHOW_COMMENT");

  /**
   * The feature that controls handling of the document
   * entity.  If this feature is present, then
   * START_ENTITY and END_ENTITY iterms are returned
   * for the document entity. 
   *
   * @see XmlItem#START_ENTITY
   * @see XmlItem#END_ENTITY
   */
  public static final XmlFeature SHOW_DOCUMENT_ENTITY
    = new BuiltinFeature("SHOW_DOCUMENT_ENTITY");

  /**
   * The feature that controls handling of references to
   * external general parsed entities.
   *
   * @see XmlItem#START_ENTITY
   * @see XmlItem#END_ENTITY
   */
  public static final XmlFeature SHOW_EXTERNAL_ENTITY
    = new BuiltinFeature("SHOW_EXTERNAL_ENTITY");

  /**
   * The feature that controls handling of references to internal
   * general parsed entities.  This does not affect references to the
   * 5 builtin entities.
   *
   * @see XmlItem#START_ENTITY
   * @see XmlItem#END_ENTITY
   */
  public static final XmlFeature SHOW_INTERNAL_ENTITY
    = new BuiltinFeature("SHOW_INTERNAL_ENTITY");

  /**
   * The feature that controls handling of unexpanded entity
   * references.
   *
   * @see XmlItem#UNEXPANDED_ENTITY_REFERENCE
   */
  public static final XmlFeature SHOW_UNEXPANDED_ENTITY_REFERENCE
    = new BuiltinFeature("SHOW_UNEXPANDED_ENTITY_REFERENCE");

  /**
   * The feature that controls handling of references to the
   * 5 builtin entities.
   *
   * @see XmlItem#BUILTIN_ENTITY_REFERENCE
   */
  public static final XmlFeature SHOW_BUILTIN_ENTITY_REFERENCE
    = new BuiltinFeature("SHOW_BUILTIN_ENTITY_REFERENCE");

  /**
   * The feature that controls handling of character references.
   *
   * @see XmlItem#CHARACTER_REFERENCE
   */
  public static final XmlFeature SHOW_CHARACTER_REFERENCE
    = new BuiltinFeature("SHOW_CHARACTER_REFERENCE");

  /**
   * The feature that controls handling of the document type
   * declaration.
   *
   * @see XmlItem#DOCUMENT_TYPE
   */
  public static final XmlFeature SHOW_DOCUMENT_TYPE
    = new BuiltinFeature("SHOW_DOCUMENT_TYPE");

  /**
   * The feature that controls handling of top-level
   * whitespace.
   *
   * @see XmlItem#TOP_LEVEL_WHITESPACE
   */
  public static final XmlFeature SHOW_TOP_LEVEL_WHITESPACE
    = new BuiltinFeature("SHOW_TOP_LEVEL_WHITESPACE");

  public static final XmlFeature ALLOW_SPLIT_TEXT
    = new BuiltinFeature("ALLOW_SPLIT_TEXT");

  public static final XmlFeature PRESERVE_ATTRIBUTE_ORDER
    = new BuiltinFeature("PRESERVE_ATTRIBUTE_ORDER");

  public static final XmlFeature SKIP_EXTERNAL_GENERAL_ENTITIES
    = new BuiltinFeature("SKIP_EXTERNAL_GENERAL_ENTITIES");

  public static final XmlFeature SKIP_PARAMETER_ENTITIES
    = new BuiltinFeature("SKIP_PARAMETER_ENTITIES");

  public static final XmlFeature SKIP_PARAMETER_ENTITIES_IF_STANDALONE
    = new BuiltinFeature("SKIP_PARAMETER_ENTITIES_IF_STANDALONE");

  /**
   * The feature that controls XML 1.0 validation.  If this feature is
   * present, then XML 1.0 validation is performed.  This feature does
   * not imply schema validation.  This feature is inconsistent with
   * the <code>SKIP_PARAMETER_ENTITIES</code>,
   * <code>SKIP_PARAMETER_ENTITIES_IF_STANDALONE</code> and
   * <code>SKIP_EXTERNAL_GENERAL_ENTITIES</code> features.
   */
  public static final XmlFeature VALIDATE = new BuiltinFeature("VALIDATE");

  static private final class UriFeature extends XmlFeature {
    private final String uri;

    private UriFeature(String uri) {
      this.uri = uri;
    }
    
    public String toString() {
      return uri;
    }
    
    public int hashCode() {
      return uri.hashCode();
    }

    public boolean equals(Object that) {
      return that instanceof UriFeature && ((UriFeature)that).uri.equals(uri);
    }
  }

  public static XmlFeature uriFeature(String uri) {
    if (uri == null)
      throw new NullPointerException();
    return new UriFeature(uri);
  }
}
