package com.thaiopensource.pullax;

import java.io.IOException;

public interface XmlItem {
  /**
   * The item type for the start of an element.
   *
   * @see #getItemType()
   */
  static final int START_ELEMENT = 1;

  /**
   * The item type for the end of an element.
   *
   * @see #getItemType()
   */
  static final int END_ELEMENT = 2;

  /**
   * The item type for characters of data.
   *
   * @see #getItemType()
   */
  static final int TEXT = 3;

  /**
   * The item type for an attribute. Attributes that serve
   * as namespace declarations are reported as attributes.
   *
   * @see #getItemType()
   */
  static final int ATTRIBUTE = 4;

  /**
   * The item type for an empty element.  An empty element is returned
   * as an <code>EMPTY_ELEMENT</code> only if the
   * <code>SHOW_EMPTY_ELEMENT</code> feature is present.  Otherwise,
   * an empty element will be returned as a
   * <code>START_ELEMENT</code>, <code>END_ELEMENT</code> pair; the
   * closing <code>&gt;</code> will be treated as belonging to the
   * <code>END_ELEMENT</code> and all the other characters will be
   * treated as belonging to the <code>START_ELEMENT</code>.
   *
   * @see #getItemType()
   */
  static final int EMPTY_ELEMENT = 5;

  /**
   * The item type for a CDATA section.  A CDATA section is returned
   * as a <code>CDATA_SECTION</code> only if the
   * <code>SHOW_CDATA_SECTION</code> feature is present.  Otherwise,
   * the characters from the CDATA section will be included in a
   * <code>TEXT</code>.
   *
   * @see #getItemType()
   */
  static final int CDATA_SECTION = 6;

  /**
   * The item type for a processing instruction. A processing
   * instruction is returned as a <code>PROCESSING_INSTRUCTION</code>
   * only if the <code>SHOW_PROCESSING_INSTRUCTION</code> feature is
   * present.  Otherwise, processing instructions are skipped.
   *
   * @see #getItemType()
   */
  static final int PROCESSING_INSTRUCTION = 7;

  /**
   * The item type for a comment.  A comment is returned as a
   * <code>COMMENT</code> only if the <code>SHOW_COMMENT</code>
   * feature is present.  Otherwise, comments are skipped.
   *
   * @see #getItemType()
   */
  static final int COMMENT = 8;

  /**
   * The item type for the start of the replacement of a general
   * entity. The context for a <code>START_ENTITY</code> entity
   * is the context of the start of the replacement text rather
   * that the context of the start of the reference. The context
   * must reflect the XML or text declaration at the beginning
   * of the entity if the entity is the document entity
   * or an external entity and the entity has an XML or text
   * declaration.  The name of the entity will be null if the
   * entity is the document entity.
   *
   * @see #getItemType()
   */
  static final int START_ENTITY = 9;

  /**
   * The item type for the end of the replacement of a general entity.
   * Context is same as context for corresponding
   * <code>START_ENTITY</code> item.
   *
   * @see #getItemType()
   */
  static final int END_ENTITY = 10;

  /**
   * The item type for an unexpanded entity reference. A reference to
   * a builtin entity is returned as a
   * <code>UNEXPANDED_ENTITY_REFERENCE</code> only if the
   * <code>SHOW_UNEXPANDED_ENTITY_REFERENCE</code> feature is
   * present. Otherwise, Otherwise, unexpanded entity references are
   * skipped.
   *
   * @see #getItemType()
   */
  static final int UNEXPANDED_ENTITY_REFERENCE = 11;

  /**
   * The item type for a reference to one of the 5 builtin entities
   * (<code>&amp;lt;</code>, <code>&amp;gt;</code>,
   * <code>&amp;amp;</code>, <code>&amp;quot;</code>,
   * <code>&amp;apos;</code>).  A reference to a builtin entity is
   * returned as a <code>BUILTIN_ENTITY_REFERENCE</code> only if the
   * <code>SHOW_BUILTIN_ENTITY_REFERENCE</code> feature is present.
   * Otherwise, the referenced character will be included in a
   * <code>TEXT</code>. The value is the referenced character.  The
   * name is the sequence of characters between the <code>&amp;</code>
   * and the <code>;</code> (exclusive).
   *
   * @see #getItemType()
   */
  static final int BUILTIN_ENTITY_REFERENCE = 12;

  /**
   * The item type for a character reference. This includes both
   * decimal and hex numeric character references but not references
   * to builtin or other entities.  A character reference is returned
   * as a <code>CHARACTER_REFERENCE</code> only if the
   * <code>SHOW_CHARACTER_REFERENCE</code> feature is present.
   * Otherwise, the referenced character will be included in a
   * <code>TEXT</code>.  The value of a
   * <code>CHARACTER_REFERENCE</code> is the referenced character
   * (which will be represented by two Java chars if the the character
   * code is greater than 0xFFFF, and otherwise by one Java char).
   * The name of the character reference is the sequence of characters
   * between the <code>&amp;#</code> and the <code>;</code>
   * (exclusive).
   *
   * @see #getItemType()
   */
  static final int CHARACTER_REFERENCE = 13;

  /**
   * The item type for a document type declaration. A document type
   * declaration is returned as a <code>DOCUMENT_TYPE</code> only
   * if the <code>SHOW_DOCUMENT_TYPE</code> feature is present.  Otherwise,
   * no item will be returned for the document type declaration
   * and its contents.
   *
   * @see #getItemType()
   */
  static final int DOCUMENT_TYPE = 14;

  /**
   * The item type for a top-level whitespace.  Top-level whitespace
   * is whitespace that matches the <a
   * href="http://www.w3.org/TR/REC-xml#NT-Misc">Misc</a> production
   * of the XML Recommendation, that is, whitespace before or after
   * the document element that is not contained within any processing
   * instruction, comment, document type declaration or XML
   * declaration.  Top-level whitespace is returned as a
   * <code>TOP_LEVEL_WHITESPACE</code> only if the
   * <code>SHOW_TOP_LEVEL_WHITESPACE</code> feature is present.
   * Otherwise, no item will be returned for top-level whitespace.
   * The value is the string of whitespace characters. It is
   * implementation dependendent how top-level whitespace is chunked
   * into <code>TOP_LEVEL_WHITESPACE</code> items.
   *
   * @see #getItemType()
   */
  static final int TOP_LEVEL_WHITESPACE = 15;

  /**
   * Returns the type of this item.
   *
   * @return the item type of this item
   *
   * @see #START_ELEMENT
   * @see #END_ELEMENT
   * @see #TEXT
   * @see #ATTRIBUTE
   * @see #EMPTY_ELEMENT
   * @see #CDATA_SECTION
   * @see #PROCESSING_INSTRUCTION
   * @see #COMMENT
   * @see #START_ENTITY
   * @see #END_ENTITY
   * @see #UNEXPANDED_ENTITY_REFERENCE
   * @see #BUILTIN_ENTITY_REFERENCE
   * @see #CHARACTER_REFERENCE
   * @see #DOCUMENT_TYPE
   * @see #TOP_LEVEL_WHITESPACE
   */
  int getItemType();

  /**
   * Returns the name of the this item, or null
   * if the item does not have a name. An item
   * has a name if and only if its item type is
   * one of <code>ATTRIBUTE</code>,
   * <code>START_ELEMENT</code>, <code>END_ELEMENT</code>,
   * <code>EMPTY_ELEMENT</code>,
   * <code>DOCUMENT_TYPE</code>,
   * <code>PROCESSING_INSTRUCTION</code>,
   * <code>START_ENTITY</code>,
   * <code>END_ENTITY</code>,
   * <code>BUILTIN_ENTITY_REFERENCE</code>,
   * <code>UNEXPANDED_ENTITY_REFERENCE</code> or
   * <code>CHARACTER_REFERENCE</code>.
   *
   * @return a String containing the name of the this item
   * or null if the item does not have a name
   */
  String getName();

  /**
   * Returns the local name of this item, or null if the item does not
   * gave a name.  The local name is the part of the name following
   * the colon. If the name does not contain a colon, then the local
   * name is the same as the name.  An item has a name if and only if
   * its item type is one of <code>ATTRIBUTE</code>,
   * <code>START_ELEMENT</code>, <code>END_ELEMENT</code>,
   * <code>EMPTY_ELEMENT</code>, <code>DOCUMENT_TYPE</code>,
   * <code>PROCESSING_INSTRUCTION</code>,
   * <code>START_ENTITY</code>,
   * <code>END_ENTITY</code>,
   * <code>BUILTIN_ENTITY_REFERENCE</code>,
   * <code>UNEXPANDED_ENTITY_REFERENCE</code> or
   * <code>CHARACTER_REFERENCE</code>.
   *
   * @return a String containing the local name of this item or
   * null if the item does not have a name
   */
  String getLocalName();

  /**
   * Returns the prefix of this item, or null if the item does not
   * have a prefix. The prefix is the part of the name preceding the
   * colon.  If the name does not contain a colon, then the prefix is
   * null. An item can have a non-null prefix only if its type is one
   * of <code>ATTRIBUTE</code>, <code>START_ELEMENT</code>,
   * <code>END_ELEMENT</code>, <code>EMPTY_ELEMENT</code>,
   * <code>DOCUMENT_TYPE</code>.
   *
   * @return a String containing the prefix of the this item or
   * null if the item does not have a prefix
   */
  String getPrefix();

  /**
   * Returns the namespace URI of this item, or null if the item does
   * not have a namespaced URI. The namespace URI of an item whose
   * prefix is <code>xml</code> is
   * <code>http://www.w3.org/XML/1998/namespace</code>. The namespace
   * URI of an attribute that is serving as a namespace declaration is
   * <code>http://www.w3.org/2000/xmlns/</code>. An item can have a
   * non-null namespace URI only if its type is one of
   * <code>ATTRIBUTE</code>, <code>START_ELEMENT</code>,
   * <code>END_ELEMENT</code> or <code>EMPTY_ELEMENT</code>.
   *
   * @return a String containing the namespace URI of this item or
   * null if the item does not have a namespace URI
   *
   * @see XmlNamespaces#XML
   * @see XmlNamespaces#XMLNS
   */
  String getNamespaceUri();

  /**
   * Returns the value of this item or null if the item does not have
   * a value.  An item has a non-null value if and only if its item
   * type is one of <code>ATTRIBUTE</code>, <code>TEXT</code>,
   * <code>PROCESSING_INSTRUCTION</code>, <code>COMMENT</code>,
   * <code>BUILTIN_ENTITY_REFERENCE</code>,
   * <code>UNEXPANDED_ENTITY_REFERENCE</code>,
   * <code>CHARACTER_REFERENCE</code> or
   * <code>TOP_LEVEL_WHITESPACE</code>.
   *
   * @return a String containing the value of the this item
   * @throws IOException if an I/O error occurs
   */
  String getValue() throws IOException;

  /**
   * Returns the location of the first character of the item.
   */
  XmlLocation getLocation();

  /**
   * Returns the context of this item.
   */
  XmlContext getContext();
}
