package com.thaiopensource.pullax.xni;

import java.io.IOException;
import java.util.Map;
import java.util.NoSuchElementException;

import org.apache.xerces.xni.XMLAttributes;

import com.thaiopensource.pullax.XmlAttributeItem;
import com.thaiopensource.pullax.XmlContext;
import com.thaiopensource.pullax.XmlContentItem;
import com.thaiopensource.pullax.XmlScanner;
import com.thaiopensource.pullax.XmlAttributeType;
import com.thaiopensource.pullax.XmlScanner;
import com.thaiopensource.pullax.XmlNamespaces;
import com.thaiopensource.pullax.XmlLocation;

class AttributeItem implements XmlAttributeItem {
  private final ItemBase parent;
  private final String name;
  private final String localName;
  private final String prefix;
  private final String namespaceUri;
  private final String value;
  private final boolean specified;
  private final boolean namespaceDecl;
  private final String type;
  private final int index;
  private final String unnormalizedValue;
  
  AttributeItem(ItemBase parent,
		XMLAttributes attributes,
		int i) {
    this.parent = parent;
    value = attributes.getValue(i);
    name = attributes.getQName(i);
    localName = attributes.getLocalName(i);
    prefix = nullify(attributes.getPrefix(i));
    if ("xmlns".equals(prefix == null ? localName : prefix)) {
      namespaceUri = XmlNamespaces.XMLNS;
      namespaceDecl = true;
      String ns = value;
      if (value.length() == 0 && prefix == null)
	ns = null;
      parent.setInternalContext(parent.getInternalContext().setNamespace(prefix, ns));
    }
    else {
      namespaceUri = nullify(attributes.getURI(i));
      namespaceDecl = false;
    }
    index = i;
    specified = attributes.isSpecified(i);
    type = attributes.getType(i);
    unnormalizedValue = attributes.getNonNormalizedValue(i);
    if (name.startsWith("xml:"))
      updateContext();
  }

  private void updateContext() {
    Context context = parent.getInternalContext();
    if (name.equals("xml:base"))
      // XXX resolve relative base URI
      context = context.setBaseUri(value);
    else if (name.equals("xml:space")) {
      XmlContext.XmlSpace xmlSpace = null;
      if (value.equals("preserve"))
	xmlSpace = XmlContext.XmlSpace.PRESERVE;
      else if (value.equals("default"))
	xmlSpace = XmlContext.XmlSpace.DEFAULT;
      else
	xmlSpace = null;
      context = context.setXmlSpace(xmlSpace);
    }
    else if (name.equals("xml:lang"))
      context = context.setXmlLang(value);
    parent.setInternalContext(context);
  }

  public XmlLocation getLocation() {
    return parent.getLocation();
  }

  public XmlContext getContext() {
    return parent.getContext();
  }

  static private String nullify(String s) {
    return "".equals(s) ? null : s;
  }

  public int getItemType() {
    return ATTRIBUTE;
  }

  public String getName() {
    return name;
  }

  public String getLocalName() {
    return localName;
  }

  public String getPrefix() {
    return prefix;
  }

  public String getNamespaceUri() {
    return namespaceUri;
  }

  public String getValue() {
    return value;
  }

  public int getAttributeIndex() {
    return index;
  }
  
  public boolean isSpecified() {
    return specified;
  }

  public boolean isNamespaceDecl() {
    return namespaceDecl;
  }

  public XmlAttributeType getAttributeType() {
    // XXX
    return null;
  }

  private class ScannerImpl implements XmlScanner {
    private final String value;
    private int index;
    private final int length;
    private final Position position;

    private ScannerImpl(String value) {
      this.value = value;
      length = value.length();
      index = 0;
      position = parent.getPosition();
    }

    public boolean hasMoreItems() {
      return index < length;
    }

    public XmlContentItem nextItem() throws IOException {
      if (index >= length)
	throw new NoSuchElementException();
      int lim = index;
    loop:
      for (lim = index; lim < length; lim++) {
	switch (value.charAt(lim)) {
	case '\n':
	case '\r':
	case '\t':
	  if (lim > index)
	    break loop;
	  index = lim + 1;
	  return makeCharacterReference(lim);
	}
      }
      String tem;
      if (index == 0 && lim == length)
	tem = value;
      else
	tem = value.substring(index, lim);
      index = lim;
      return new TextItem(parent.getInternalContext(), position, tem);
    }

    private XmlContentItem makeCharacterReference(int i) {
      String s = value.substring(i, i + 1);
      char c = value.charAt(i);
      String name;
      if (c == '\n')
	name = "xA";
      else if (c == '\r')
	name = "xD";
      else
	name = "9";
      return new CharacterReferenceItem(parent.getInternalContext(),
					position,
					name,
					s);
    }

    public void close() { }
  }

  public XmlScanner scanUnnormalizedValue() {
    return new ScannerImpl(value);
  }
}
