package com.thaiopensource.pullax.xni;

import com.thaiopensource.pullax.XmlElementType;
import com.thaiopensource.pullax.XmlExternalId;
import com.thaiopensource.pullax.XmlAttributeDefinition;
import com.thaiopensource.pullax.XmlAttributeType;
import com.thaiopensource.pullax.XmlEntity;

import org.apache.xerces.xni.XMLString;
import org.apache.xerces.xni.XMLDTDHandler;
import org.apache.xerces.xni.XMLLocator;

import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import java.util.Iterator;

class DtdBuilder implements XMLDTDHandler {
  private Dtd dtd;
  private StringBuffer internalSubset = new StringBuffer();
  private boolean inInternalSubset = true;
  private EndDtdHandler handler;
  private Position startPosition;
  private Position position;
  private String rootElement;
  private Context context;
  private XMLLocator locator;
  private ElementType currentElementType = null;

  private static class ElementType implements XmlElementType {
    private boolean elementContent = false;
    private boolean defined = false;

    private Map attributeDefinitions = new HashMap();

    private void setElementContent(boolean b) {
      if (defined)
	return;
      defined = true;
      elementContent = b;
    }

    public boolean hasElementContent() {
      return elementContent;
    }
    
    public Map getAttributeDefinitions() {
      return attributeDefinitions;
    }

    private void freeze() {
      attributeDefinitions
	= Collections.unmodifiableMap(attributeDefinitions);
    }
  }

  private static class AttributeDefinition implements XmlAttributeDefinition {
    private XmlAttributeType type;
    private String defaultValue;

    private AttributeDefinition(XmlAttributeType type) {
      this.type = type;
    }
    public XmlAttributeType getAttributeType() {
      return type;
    }
    public String getDefaultValue() {
      return defaultValue;
    }
  }

  private static class Entity implements XmlEntity {
    private final String notation;
    private final String replacementText;
    private final XmlExternalId externalId;
    
    private Entity(XmlExternalId externalId) {
      this.replacementText = null;
      this.externalId = externalId;
      this.notation = null;
    }

    private Entity(XmlExternalId externalId, String notation) {
      this.replacementText = null;
      this.externalId = externalId;
      this.notation = notation;
    }

    private Entity(String replacementText) {
      this.replacementText = replacementText;
      this.externalId = null;
      this.notation = null;
    }

    public XmlExternalId getExternalId() {
      return externalId;
    }
	
    public String getReplacementText() {
      return replacementText;
    }

    public String getNotation() {
      return notation;
    }
  }

  void prepare(EndDtdHandler handler,
	       Context context,
	       Position position,
	       String rootElement,
	       String publicId,
	       String systemId) {
    this.handler = handler;
    this.dtd = new Dtd(new XmlExternalId(publicId, systemId, null));
    this.rootElement = rootElement;
    this.startPosition = new Position(position);
    this.position = position;
    this.context = context;
    internalSubset.append("\n");
  }

  public void startDTD(XMLLocator locator) {
    this.locator = locator;
  }
    
  public void startEntity(String name, 
			  String publicId, String systemId,
			  String baseSystemId,
			  String encoding) {
    if (name.equals("[dtd]"))
      inInternalSubset = false;
  }
    
  public void textDecl(String version, String encoding) {
  }
    
  public void endEntity(String name) {
  }
    
  public void comment(XMLString text) {
    if (!inInternalSubset)
      return;
    internalSubset.append("<!--");
    internalSubset.append(text.ch, text.offset, text.length);
    internalSubset.append("-->\n");
    updatePosition();
  }
    
  public void processingInstruction(String target, XMLString data) {
    context.freeze();
    // XXX need to handle positions the same was as in XmlScanner
    dtd.addProcessingInstruction(new ProcessingInstructionItem(context, position,
							       target, data.toString()));
    if (!inInternalSubset)
      return;
    internalSubset.append("<?");
    internalSubset.append(target);
    if (data.length > 0) {
      internalSubset.append(" ");
      internalSubset.append(data.ch, data.offset, data.length);
    }
    internalSubset.append("?>\n");
    updatePosition();
  }
  
  private ElementType lookupElementType(String name) {
    Map elementTypes = dtd.getElementTypes();
    ElementType elementType = (ElementType)elementTypes.get(name);
    if (elementType == null) {
      elementType = new ElementType();
      elementTypes.put(name, elementType);
    }
    return elementType;
  }

  private static boolean isElementContent(String contentModel) {
    return contentModel.charAt(0) == '(' && contentModel.charAt(1) != '#';
  }

  public void elementDecl(String name, String contentModel) {
    lookupElementType(name).setElementContent(isElementContent(contentModel));
    if (!inInternalSubset)
      return;
    internalSubset.append("<!ELEMENT ");
    internalSubset.append(name);
    internalSubset.append(" ");
    internalSubset.append(contentModel);
    internalSubset.append(">\n");
    updatePosition();
  }
    
  public void startAttlist(String elementName) {
    if (!inInternalSubset)
      return;
    internalSubset.append("<!ATTLIST ");
    internalSubset.append(elementName);
    updatePosition();
  }

  public void attributeDecl(String elementName, String attributeName, 
			    String type, String[] enumeration, 
			    String defaultType, XMLString defaultValue) {
    // Work around Xerces beta 3 bug
    // #IMPLIED, #REQUIRED should have defaultValue == null
    if (defaultType != null && !defaultType.equals("#FIXED"))
      defaultValue = null;
    Map attDefs = lookupElementType(elementName).getAttributeDefinitions();
    if (attDefs.get(attributeName) == null) {
      AttributeDefinition def
	= new AttributeDefinition(XmlAttributeType.valueOf(type));
      if (defaultValue != null)
	def.defaultValue = defaultValue.toString();
      attDefs.put(attributeName, def);
    }
    if (!inInternalSubset)
      return;
    internalSubset.append("\n  ");
    internalSubset.append(attributeName);
    if (!type.equals("ENUMERATION")) {
      internalSubset.append(' ');
      internalSubset.append(type);
    }
    if (enumeration != null) {
      internalSubset.append(" (");
      internalSubset.append(enumeration[0]);
      for (int i = 1; i < enumeration.length; i++) {
	internalSubset.append('|');
	internalSubset.append(enumeration[i]);
      }
      internalSubset.append(')');
    }
    if (defaultType != null) {
      internalSubset.append(' ');
      internalSubset.append(defaultType);
    }
    if (defaultValue != null) {
      internalSubset.append(" \"");
      for (int i = defaultValue.offset, n = defaultValue.length;
	   n > 0;
	   n--, i++) {
	char ch = defaultValue.ch[i];
	switch (ch) {
	case '<':
	  internalSubset.append("&lt;");
	  break;
	case '"':
	  internalSubset.append("&quot;");
	  break;
	case '&':
	  internalSubset.append("&amp;");
	  break;
	case '\n':
	  internalSubset.append("&#xA;");
	  break;
	case '\r':
	  internalSubset.append("&#xD;");
	  break;
	case '\t':
	  internalSubset.append("&#x9;");
	  break;
	default:
	  internalSubset.append(ch);
	  break;
	}
      }
      internalSubset.append('"');
    }
    updatePosition();
  }
    
  public void endAttlist() {
    if (!inInternalSubset)
      return;
    internalSubset.append(">\n");
    updatePosition();
  }
    
  public void internalEntityDecl(String name, XMLString text, 
				 XMLString nonNormalizedText) 
  {
    if (name.charAt(0) != '%') {
      Map entities = dtd.getGeneralEntities();
      if (entities.get(name) == null)
	entities.put(name, new Entity(text.toString()));
    }
    if (!inInternalSubset)
      return;
    internalSubset.append("<!ENTITY ");
    entityName(name);
    internalSubset.append(' ');
    literal(nonNormalizedText.toString());
    internalSubset.append(">\n");
    updatePosition();
  }
    
  public void externalEntityDecl(String name, 
				 String publicId, String systemId,
				 String baseSystemId) {
    if (name.charAt(0) != '%') {
      Map entities = dtd.getGeneralEntities();
      if (entities.get(name) == null)
	entities.put(name, new Entity(new XmlExternalId(publicId,
							systemId,
							baseSystemId)));
    }
    if (!inInternalSubset)
      return;
    internalSubset.append("<!ENTITY ");
    entityName(name);
    externalId(publicId, systemId);
    internalSubset.append(">\n");
    updatePosition();
  }

  private void entityName(String name) {
    if (name.charAt(0) == '%') {
      internalSubset.append("% ");
      internalSubset.append(name.substring(1));
    }
    else
      internalSubset.append(name);
  }

  private void externalId(String publicId, String systemId) {
    if (publicId != null) {
      internalSubset.append(" PUBLIC ");
      literal(publicId);
    }
    if (systemId != null) {
      internalSubset.append(" SYSTEM ");
      literal(systemId);
    }
  }

  private void literal(String s) {
    char lit = s.indexOf('"') >= 0 ? '\'' : '"';
    internalSubset.append(lit);
    internalSubset.append(s);
    internalSubset.append(lit);
  }

  public void unparsedEntityDecl(String name, 
				 String publicId, String systemId, 
				 String notation) {
    Map entities = dtd.getGeneralEntities();
    if (entities.get(name) == null)
      entities.put(name,
		   new Entity(new XmlExternalId(publicId,
						systemId,
						null),
			      notation));
    if (!inInternalSubset)
      return;
    internalSubset.append("<!ENTITY");
    internalSubset.append(' ');
    internalSubset.append(name);
    externalId(publicId, systemId);
    internalSubset.append(' ');
    internalSubset.append(notation);
    internalSubset.append(">\n");
    updatePosition();
  }
    
  public void notationDecl(String name, String publicId, String systemId) {
    Map notations = dtd.getNotations();
    if (notations.get(name) == null)
      notations.put(name, new XmlExternalId(publicId, systemId, null));
    if (!inInternalSubset)
      return;
    internalSubset.append("<!NOTATION ");
    internalSubset.append(name);
    externalId(publicId, systemId);
    internalSubset.append(">\n");
    updatePosition();
  }
    
  public void startConditional(short type) {
  }
    
  public void characters(XMLString text) {
  }
    
  public void endConditional() {
  }
    
  public void endDTD() {
    if (internalSubset.length() > 1)
      dtd.setInternalSubset(internalSubset.toString());
    for (Iterator iter = dtd.getElementTypes().values().iterator();
	 iter.hasNext();)
      ((ElementType)iter.next()).freeze();
    dtd.freeze();
    context = context.setDtd(dtd);
    handler.endDtd(context,
		   new DocumentTypeItem(context,
					startPosition,
					rootElement));
    updatePosition();
  }

  private final void updatePosition() {
    position.lineNumber = locator.getLineNumber();
    position.columnNumber = locator.getColumnNumber();
  }

}
