package com.thaiopensource.pullax.xni;

import java.io.IOException;
import java.util.NoSuchElementException;
import java.util.Locale;

import com.thaiopensource.pullax.XmlScanner;
import com.thaiopensource.pullax.XmlContentItem;
import com.thaiopensource.pullax.XmlAttributeItem;
import com.thaiopensource.pullax.XmlItem;
import com.thaiopensource.pullax.XmlContext;
import com.thaiopensource.pullax.XmlLocation;
import com.thaiopensource.pullax.XmlInputSource;
import com.thaiopensource.pullax.XmlReaderInputSource;
import com.thaiopensource.pullax.XmlInputStreamInputSource;
import com.thaiopensource.pullax.XmlExternalId;
import com.thaiopensource.pullax.XmlReporter;
import com.thaiopensource.pullax.XmlReport;
import com.thaiopensource.pullax.XmlReportException;
import com.thaiopensource.pullax.XmlEntityResolver;
import com.thaiopensource.pullax.XmlAbortException;
import com.thaiopensource.pullax.XmlSeverity;

import org.apache.xerces.xni.parser.XMLPullParserConfiguration;
import org.apache.xerces.xni.parser.XMLInputSource;
import org.apache.xerces.xni.parser.XMLConfigurationException;
import org.apache.xerces.xni.parser.XMLErrorHandler;
import org.apache.xerces.xni.parser.XMLParseException;
import org.apache.xerces.xni.XMLDocumentHandler;
import org.apache.xerces.xni.XNIException;
import org.apache.xerces.xni.XMLString;
import org.apache.xerces.xni.XMLLocator;
import org.apache.xerces.xni.XMLAttributes;
import org.apache.xerces.xni.QName;

import org.apache.xerces.impl.Constants;
import org.apache.xerces.impl.XMLEntityManager;

class XmlScannerImpl extends FeatureSet implements XmlScanner {
  private final XMLPullParserConfiguration parser;
  private final XmlEntityResolver resolver;
  private XmlContentItem[] queue = new XmlContentItem[2];
  private int queueSize = 0;
  private int queueIndex = 0;
  private boolean completed = false;

  private DtdBuilder dtdBuilder = new DtdBuilder();

  private static class EntityEntry {
    EntityEntry() {
      prev = null;
    }
    EntityEntry(EntityEntry prev) {
      this.prev = prev;
    }
    Context context;
    Position position;
    int internalCount;
    EntityEntry prev;
  }

  private static class ElementEntry {
    private final Context context;
    private final int dupContext;
    private final ElementEntry parent;

    ElementEntry(Context context, int dupContext, ElementEntry parent) {
      this.context = context;
      this.dupContext = dupContext;
      this.parent = parent;
    }
  }

  static private final int NOT_BUFFERING = 0;
  static private final int BUFFERING_CDATA = 1;
  static private final int BUFFERING_TEXT = 2;
  static private final int BUFFERING_CHARACTER_REFERENCE = 3;
  static private final int BUFFERING_BUILTIN_ENTITY_REFERENCE = 4;

  private class DocumentHandlerImpl implements XMLDocumentHandler, EndDtdHandler {
    private StringBuffer buf = new StringBuffer();
    private Context context = new Context();
    private Position position = new Position();
    private XMLLocator locator;
    private EntityEntry entityStack = new EntityEntry();
    private ElementEntry elementStack;
    private int dupContext = 0;
    private int buffering = NOT_BUFFERING;
    private TextItem pendingTextItem = null;
    private boolean pendingStartEntity = showDocumentEntity;
    private String pendingStartEntityName = null;

    public void startDocument(XMLLocator locator,
			      String encoding)
      throws XNIException {
      this.locator = locator;
      context = context.setCharacterEncodingScheme(encoding);
      String uri = XMLEntityManager.expandSystemId(locator.getSystemId(),
						   locator.getBaseSystemId());
      context = context.setSourceUri(uri);
      context = context.setBaseUri(uri);
      updatePosition();
    }

    private final void updatePosition() {
      position.lineNumber = locator.getLineNumber();
      position.columnNumber = locator.getColumnNumber();
    }

    public void xmlDecl(String version,
			String encoding,
			String standalone)
      throws XNIException {
      context = context.setVersion(version);
      context = context.setCharacterEncodingScheme(encoding);
      context = context.setStandalone(toStandalone(standalone));
      entityStarted();
      updatePosition();
    }

    private void entityStarted() {
      if (pendingStartEntity) {
	context.freeze();
	append(new StartEntityItem(context, position, pendingStartEntityName));
	pendingStartEntity = false;
      }
    }
      
    public void doctypeDecl(String rootElement,
			    String publicId,
			    String systemId)
      throws XNIException {
      entityStarted();
      dtdBuilder.prepare(this, context, position,
			 rootElement, publicId, systemId);
      updatePosition();
    }
    
    public void endDtd(Context context, XmlContentItem item) {
      this.context = context;
      if (showDocumentType)
	append(item);
    }

    public void comment(XMLString text)
      throws XNIException {
      entityStarted();
      if (showComment) {
	if (pendingTextItem != null)
	  flushText();
	context.freeze();
	append(new CommentItem(context, position, text.toString()));
      }
      updatePosition();
    }

    public void processingInstruction(String target,
				      XMLString data)
      throws XNIException {
      entityStarted();
      if (showProcessingInstruction) {
	if (pendingTextItem != null)
	  flushText();
	context.freeze();
	append(new ProcessingInstructionItem(context, position,
					     target, data.toString()));
      }
      updatePosition();
    }

    public void startPrefixMapping(String prefix,
				   String uri)
      throws XNIException {
    }

    public void startElement(QName element,
			     XMLAttributes attributes)
      throws XNIException {
      entityStarted();
      if (pendingTextItem != null)
	flushText();
      context.freeze();
      ItemBase item = new StartElementItem(context, position,
					   element, attributes);
      // The context may be changed by the attributes.
      Context newContext = item.getInternalContext();
      if (newContext == context)
	dupContext++;
      else {
	elementStack = new ElementEntry(context, dupContext, elementStack);
	context = newContext;
	context.freeze();
	dupContext = 0;
      }
      append(item);
      updatePosition();
    }

    public void emptyElement(QName element,
			     XMLAttributes attributes)
      throws XNIException {
      entityStarted();
      if (pendingTextItem != null)
	flushText();
      context.freeze();
      ItemBase item;
      if (showEmptyElement)
	item = new EmptyElementItem(context, position, element, attributes);
      else
	item = new StartElementItem(context, position, element, attributes);
      // The context may be changed by the attributes.
      Context newContext = item.getInternalContext();
      newContext.freeze();
      append(item);
      updatePosition();
      if (!showEmptyElement)
	append(new EndElementItem(newContext, position, element));
    }

    public void endElement(QName element) throws XNIException {
      if (pendingTextItem != null)
	flushText();
      context.freeze();
      append(new EndElementItem(context, position, element));
      if (dupContext > 0)
	--dupContext;
      else {
	context = elementStack.context;
	dupContext = elementStack.dupContext;
	elementStack = elementStack.parent;
      }
      updatePosition();
    }

    public void startEntity(String name,
			    String publicId,
			    String systemId,
			    String baseSystemId,
			    String encoding)
      throws XNIException {
      entityStarted();
      if (showCharacterReference && name.charAt(0) == '#') {
	if (pendingTextItem != null)
	  flushText();
	buffering = BUFFERING_CHARACTER_REFERENCE;
      }
      else if (showBuiltinEntityReference && isBuiltinEntity(name)) {
	if (pendingTextItem != null)
	  flushText();
	buffering = BUFFERING_BUILTIN_ENTITY_REFERENCE;
      }
      else {
	if (systemId != null) {
	  if (showExternalEntity) {
	    if (pendingTextItem != null)
	      flushText();
	    pendingStartEntity = true;
	    pendingStartEntityName = name;
	  }
	  entityStack = new EntityEntry(entityStack);
	  entityStack.context = context;
	  entityStack.position = position;
	  position = new Position();
	  String uri = XMLEntityManager.expandSystemId(systemId,
						       baseSystemId);
	  context = context.setSourceUri(uri);
	  context = context.setBaseUri(uri);
	  context = context.setCharacterEncodingScheme(encoding);
	}
	else {
	  if (showInternalEntity) {
	    if (pendingTextItem != null)
	      flushText();
	    context.freeze();
	    append(new StartEntityItem(context, position, name));
	  }
	  entityStack.internalCount++;
	}
	updatePosition();
      }
    }

    public void endEntity(String name) throws XNIException {
      entityStarted();
      if (buffering == BUFFERING_CHARACTER_REFERENCE
	  || buffering == BUFFERING_BUILTIN_ENTITY_REFERENCE) {
	context.freeze();
	XmlContentItem item;
	if (buffering == BUFFERING_BUILTIN_ENTITY_REFERENCE)
	  item = new BuiltinEntityReferenceItem(context,
						position,
						name,
						buf.toString());
	else
	  item = new CharacterReferenceItem(context,
					    position,
					    name.substring(1),
					    buf.toString());
	append(item);
	buf.setLength(0);
	buffering = NOT_BUFFERING;
	updatePosition();
      }
      else {
	if (entityStack.internalCount > 0) {
	  if (showInternalEntity) {
	    if (pendingTextItem != null)
	      flushText();
	    context.freeze();
	    append(new EndEntityItem(context, position, name));
	  }
	  entityStack.internalCount--;
	}
	else {
	  if (showExternalEntity) {
	    if (pendingTextItem != null)
	      flushText();
	    context.freeze();
	    append(new EndEntityItem(context, position, name));
	  }
	  position = entityStack.position;
	  position.columnNumber += name.length() + 2;
	  context = entityStack.context;
	  entityStack = entityStack.prev;
	}
      }
    }

    public void textDecl(String version,
			 String encoding)
      throws XNIException {
      context = context.setVersion(version);
      context = context.setCharacterEncodingScheme(encoding);
      entityStarted();
      updatePosition();
    }

    public void characters(XMLString text) throws XNIException {
      entityStarted();
      if (buffering != NOT_BUFFERING) {
	buf.append(text.ch, text.offset, text.length);
	if (buffering == BUFFERING_TEXT)
	  updatePosition();
      }
      else if (allowSplitText) {
	context.freeze();
	append(new TextItem(context, position, text.toString()));
	updatePosition();
      }
      else {
	context.freeze();
	pendingTextItem = new TextItem(context, position);
	buf.append(text.ch, text.offset, text.length);
	buffering = BUFFERING_TEXT;
	updatePosition();
      }
    }

    private void flushText() {
      pendingTextItem.setValue(buf.toString());
      buf.setLength(0);
      append(pendingTextItem);
      pendingTextItem = null;
      buffering = NOT_BUFFERING;
    }

    public void ignorableWhitespace(XMLString text) throws XNIException {
      characters(text);
    }

    public void endPrefixMapping(String prefix) throws XNIException {
    }

    public void startCDATA() throws XNIException {
      entityStarted();
      if (showCdataSection) {
	if (pendingTextItem != null)
	  flushText();
	buffering = BUFFERING_CDATA;
      }
      else
	updatePosition();
    }

    public void endCDATA() throws XNIException {
      if (buffering == BUFFERING_CDATA) {
	context.freeze();
	append(new CdataSectionItem(context, position, buf.toString()));
	buf.setLength(0);
	buffering = NOT_BUFFERING;
      }
      updatePosition();
    }

    public void endDocument() throws XNIException {
      updatePosition();
      if (showDocumentEntity) {
	context.freeze();
	append(new EndEntityItem(context, position, null));
      }
    }
  }

  private static final String FEATURE_NOTIFY_CHAR_REFS =
    Constants.XERCES_FEATURE_PREFIX + Constants.NOTIFY_CHAR_REFS_FEATURE;

  private static final String FEATURE_NOTIFY_BUILTIN_REFS =
    Constants.XERCES_FEATURE_PREFIX + Constants.NOTIFY_BUILTIN_REFS_FEATURE;

  static private class ErrorHandler implements XMLErrorHandler {
    private final XmlReporter reporter;
    ErrorHandler(XmlReporter reporter) {
      this.reporter = reporter;
    }

    public void warning(String domain, String key, 
                        XMLParseException exception) {
      try {
	reporter.report(new XmlReport(XmlSeverity.WARNING,
				      new Location(exception),
				      exception.getMessage()));
      }
      catch (IOException e) {
	throw new XNIException(e);
      }
    }

    public void error(String domain, String key, 
                        XMLParseException exception) {
      try {
	reporter.report(new XmlReport(XmlSeverity.ERROR,
				      new Location(exception),
				      exception.getMessage()));
      }
      catch (IOException e) {
	throw new XNIException(e);
      }
    }

    public void fatalError(String domain, String key, 
			   XMLParseException exception) {
      try {
	reporter.report(new XmlReport(XmlSeverity.FATAL_ERROR,
				      new Location(exception),
				      exception.getMessage()));
      }
      catch (IOException e) {
	throw new XNIException(e);
      }
      throw new XNIException(new XmlAbortException());
    }
  }

  static private class Location implements XmlLocation {
    private final XMLParseException exception;
    Location(XMLParseException exception) {
      this.exception = exception;
    }
    public String getUri() {
      return exception.getSystemId();
    }
    public int getLineNumber() {
      return exception.getLineNumber();
    }
    public int getColumnNumber() {
      return exception.getColumnNumber();
    }
  }

  static private class Reporter implements XmlReporter {
    public void report(XmlReport report) throws XmlReportException {
      if (report.getSeverity() == XmlSeverity.FATAL_ERROR
	  || report.getSeverity() == XmlSeverity.ERROR)
	throw new XmlReportException(report);
    }
  }

  XmlScannerImpl(XMLPullParserConfiguration parser,
		 XmlInputSource input,
		 FeatureSet fs,
		 Locale locale,
		 XmlEntityResolver resolver,
		 XmlReporter reporter) throws IOException {
    super(fs);
    this.parser = parser;
    this.resolver = resolver;
    parser.setDocumentHandler(new DocumentHandlerImpl());
    if (reporter == null)
      reporter = new Reporter();
    parser.setErrorHandler(new ErrorHandler(reporter));
    parser.setDTDHandler(dtdBuilder);
    try {
      parser.setLocale(locale);
    }
    catch (XNIException e) {
    }
    if (showCharacterReference) {
      try {
	parser.setFeature(FEATURE_NOTIFY_CHAR_REFS, true);
      }
      catch (XMLConfigurationException e) {
	throw new RuntimeException("could not configure char refs");
      }
    }
    if (showBuiltinEntityReference) {
      try {
	parser.setFeature(FEATURE_NOTIFY_BUILTIN_REFS, true);
      }
      catch (XMLConfigurationException e) {
	throw new RuntimeException("could not configure builtin refs");
      }
    }
    parser.setInputSource(makeInputSource(input));
  }

  private static XMLInputSource makeInputSource(XmlInputSource obj) {
    if (obj instanceof XmlReaderInputSource) {
      XmlReaderInputSource input = ((XmlReaderInputSource)obj);
      return new XMLInputSource(null,
				input.getUri(),
				input.getUri(),
				input.getReader(),
				input.getEncoding());
    }
    if (obj instanceof XmlInputStreamInputSource) {
      XmlInputStreamInputSource input = ((XmlInputStreamInputSource)obj);
      return new XMLInputSource(null,
				input.getUri(),
				input.getUri(),
				input.getInputStream(),
				input.getEncoding());
    }
    XmlExternalId input = (XmlExternalId)obj;
    return new XMLInputSource(input.getPublicId(),
			      input.getSystemId(),
			      input.getBaseUri());
  }

  public boolean hasMoreItems() throws IOException {
    try {
      while (queueSize == 0) {
	if (completed)
	  return false;
	if (!parser.parse(false))
	  completed = true;
      }
      return true;
    }
    catch (XNIException e) {
      handleXNIException(e);
      throw e;
    }
  }

  public void handleXNIException(XNIException e) throws IOException {
    if (e instanceof XMLParseException)
      throw new XmlReportException(new XmlReport(XmlSeverity.FATAL_ERROR,
						 new Location((XMLParseException)e),
						 e.getMessage()));
    Exception cause = e.getException();
    if (cause instanceof RuntimeException)
      throw (RuntimeException)cause;
    if (cause instanceof IOException)
      throw (IOException)cause;
  }

  public XmlContentItem nextItem() throws IOException {
    while (queueSize == 0) {
      if (completed)
	throw new NoSuchElementException();
      try {
	if (!parser.parse(false))
	  completed = true;
      }
      catch (XNIException e) {
	handleXNIException(e);
	throw e;
      }
    }
    XmlContentItem tem = queue[queueIndex];
    queueIndex = (queueIndex + 1) % queue.length;
    queueSize--;
    return tem;
  }

  private final void append(XmlContentItem item) {
    if (queueSize == queue.length) {
      XmlContentItem[] newQueue = new XmlContentItem[queue.length << 1];
      System.arraycopy(queue, queueIndex,
		       newQueue, 0,
		       queue.length - queueIndex);
      System.arraycopy(queue, 0,
		       newQueue, queue.length - queueIndex,
		       queueIndex);
      queue = newQueue;
      queueIndex = 0;
    }
    queue[(queueIndex + queueSize++) % queue.length] = item;
  }

  public void close() throws IOException {
  }

  private static XmlContext.Standalone toStandalone(String s) {
    if (s == null)
      return null;
    if (s.equals("yes"))
      return XmlContext.Standalone.YES;
    return XmlContext.Standalone.NO;
  }

  private static boolean isBuiltinEntity(String s) {
    switch (s.charAt(0)) {
    case 'a':
      return s.equals("amp") || s.equals("apos");
    case 'l':
      return s.equals("lt");
    case 'g':
      return s.equals("gt");
    case 'q':
      return s.equals("quot");
    }
    return false;
  }
      
}
